
// p4fliconv: high resolution interlaced FLI converter utility
// Copyright (C) 2007-2008 Istvan Varga <istvanv@users.sourceforge.net>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// The Plus/4 program files generated by this utility are not covered by the
// GNU General Public License, and can be used, modified, and distributed
// without any restrictions.

#include "plus4emu.hpp"
#include "decompress.hpp"
#include <vector>

namespace Plus4FLIConv {

  unsigned int PRGDecompressor::readBits(size_t nBits)
  {
    unsigned int  retval = 0U;
    for (size_t i = 0; i < nBits; i++) {
      if (shiftRegisterCnt < 1) {
        if (inputBufferPosition >= inputBufferSize)
          throw Plus4Emu::Exception("unexpected end of compressed data");
        shiftRegister = inputBuffer[inputBufferPosition];
        shiftRegisterCnt = 8;
        inputBufferPosition++;
      }
      retval = (retval << 1) | (unsigned int) ((shiftRegister >> 7) & 0x01);
      shiftRegister = (shiftRegister & 0x7F) << 1;
      shiftRegisterCnt--;
    }
    return retval;
  }

  unsigned int PRGDecompressor::readLZMatchParameterBits(unsigned char n)
  {
    if (n < 0x08)
      return (unsigned int) n;
    if (n >= 0x3C)
      throw Plus4Emu::Exception("error in compressed data");
    unsigned int  retval = (unsigned int) ((n & 0x03) | 0x04);
    unsigned int  nBits = (unsigned int) (n >> 2) - 1U;
    retval = (retval << nBits) | readBits(nBits);
    return retval;
  }

  unsigned int PRGDecompressor::gammaDecode()
  {
    unsigned int  retval = 1U;
    unsigned int  cnt = 1U;
    while (true) {
      if (readBits(1) == 0U)
        break;
      retval = (retval << 1) | readBits(1);
      cnt++;
      if (cnt > 16U)
        throw Plus4Emu::Exception("error in compressed data");
    }
    return retval;
  }

  unsigned int PRGDecompressor::huffmanDecode(int huffTable)
  {
    if (huffTable == 0 && !usingHuffTable0)
      return readBits(9);
    if (huffTable != 0 && !usingHuffTable1)
      return readBits(5);
    unsigned int  tmp = 1U;
    unsigned int  cnt = 0U;
    const unsigned int  *limitTable = huffmanLimitTable0;
    const unsigned int  *offsetTable = huffmanOffsetTable0;
    const unsigned int  *decodeTable = huffmanDecodeTable0;
    if (huffTable != 0) {
      limitTable = huffmanLimitTable1;
      offsetTable = huffmanOffsetTable1;
      decodeTable = huffmanDecodeTable1;
    }
    while (true) {
      if (cnt >= 16U)
        throw Plus4Emu::Exception("error in compressed data");
      tmp = (tmp << 1) | readBits(1);
      if (tmp < limitTable[cnt])
        break;
      cnt++;
    }
    tmp = (tmp + offsetTable[cnt]) & 0xFFFFU;
    if (huffTable == 0 && tmp >= 324U)
      throw Plus4Emu::Exception("error in compressed data");
    if (huffTable != 0 && tmp >= 28U)
      throw Plus4Emu::Exception("error in compressed data");
    return decodeTable[tmp];
  }

  unsigned char PRGDecompressor::readDeltaValue()
  {
    unsigned char retval = (unsigned char) readBits(7);
    if (retval < 0x40)
      retval = retval + 0xC0;
    else
      retval = retval - 0x3F;
    return retval;
  }

  unsigned int PRGDecompressor::readMatchLength()
  {
    unsigned char n = 0x00;
    if (usingHuffTable1)
      n = (unsigned char) huffmanDecode(1);
    else
      n = (unsigned char) readBits(5);
    unsigned int  retval = readLZMatchParameterBits(n);
    if (retval > 254U)
      throw Plus4Emu::Exception("error in compressed data");
    retval = retval + 2U;
    return retval;
  }

  void PRGDecompressor::huffmanInit(int huffTable)
  {
    unsigned int  *limitTable = huffmanLimitTable0;
    unsigned int  *offsetTable = huffmanOffsetTable0;
    unsigned int  *decodeTable = huffmanDecodeTable0;
    size_t        nSymbols = 324;
    if (huffTable != 0) {
      limitTable = huffmanLimitTable1;
      offsetTable = huffmanOffsetTable1;
      decodeTable = huffmanDecodeTable1;
      nSymbols = 28;
    }
    // clear tables
    for (size_t i = 0; i < 16; i++) {
      limitTable[i] = 0U;
      offsetTable[i] = 0xFFFFFFFFU;
    }
    for (size_t i = 0; i < nSymbols; i++)
      decodeTable[i] = 0xFFFFFFFFU;
    if (readBits(1) == 0U) {
      // Huffman coding is disabled, nothing to do
      if (huffTable == 0)
        usingHuffTable0 = false;
      else
        usingHuffTable1 = false;
      return;
    }
    if (huffTable == 0)
      usingHuffTable0 = true;
    else
      usingHuffTable1 = true;
    unsigned int  tmp = 1U;
    size_t        tablePos = 0;
    bool          doneFlag = false;
    for (size_t i = 0; i < 16; i++) {
      tmp = tmp << 1;
      offsetTable[i] = tablePos - tmp;
      size_t  cnt = gammaDecode() - 1U;
      if (cnt > nSymbols)
        throw Plus4Emu::Exception("error in compressed data");
      unsigned int  decodedValue = 0U - 1U;
      for (size_t j = 0; j < cnt; j++) {
        if (doneFlag)
          throw Plus4Emu::Exception("error in compressed data");
        decodedValue = decodedValue + gammaDecode();
        if (decodedValue >= (unsigned int) nSymbols || tablePos >= nSymbols)
          throw Plus4Emu::Exception("error in compressed data");
        if (tmp == 0xFFU)       // 1111111 is never output by the compressor
          throw Plus4Emu::Exception("error in compressed data");
        decodeTable[tablePos] = decodedValue;
        tablePos++;
        tmp++;
        if (tmp >= (1U << (i + 2)))
          doneFlag = true;      // overflow (only allowed for the last symbol)
      }
      limitTable[i] = tmp;
    }
  }

  bool PRGDecompressor::decompressDataBlock(std::vector< unsigned char >& buf,
                                            std::vector< bool >& bytesUsed)
  {
    unsigned int  startAddr = readBits(16);
    unsigned int  nBytes = (readBits(16) ^ 0xFFFFU) + 1U;
    bool          isLastBlock = readBits(1);
    bool          compressionEnabled = readBits(1);
    if (!compressionEnabled) {
      // copy uncompressed data
      for (unsigned int i = 0U; i < nBytes; i++) {
        if (startAddr > 0xFFFFU || bytesUsed[startAddr])
          throw Plus4Emu::Exception("error in compressed data");
        buf[startAddr] = (unsigned char) readBits(8);
        bytesUsed[startAddr] = true;
        startAddr++;
      }
      return isLastBlock;
    }
    huffmanInit(0);
    huffmanInit(1);
    prvMatchOffsetsPos = 0U;
    for (int i = 0; i < 4; i++)         // clear recent match offsets buffer
      prvMatchOffsets[i] = 0xFFFFFFFFU;
    for (unsigned int i = 0U; i < nBytes; ) {
      unsigned int  tmp = 0U;
      if (usingHuffTable0)
        tmp = huffmanDecode(0);
      else
        tmp = readBits(9);
      if (tmp >= 0x0144U)
        throw Plus4Emu::Exception("error in compressed data");
      if (tmp <= 0xFFU) {
        // literal character
        if (startAddr > 0xFFFFU || bytesUsed[startAddr])
          throw Plus4Emu::Exception("error in compressed data");
        buf[startAddr] = (unsigned char) tmp;
        bytesUsed[startAddr] = true;
        startAddr++;
        i++;
        continue;
      }
      unsigned int  offs = 0U;
      unsigned char deltaValue = 0x00;
      if (tmp >= 0x013CU) {
        // check for special match codes:
        if (tmp < 0x0140U) {
          // match with delta value
          deltaValue = readDeltaValue();
          offs = tmp & 0x03U;
        }
        else {
          // use recent match offset
          offs = prvMatchOffsets[(prvMatchOffsetsPos + tmp) & 3U];
        }
      }
      else if (tmp < 0x0108) {
        // short match offset
        offs = tmp & 0x07U;
      }
      else {
        offs = readLZMatchParameterBits((unsigned char) (tmp & 0xFFU));
        // store in recent offsets buffer
        prvMatchOffsetsPos = (prvMatchOffsetsPos - 1U) & 3U;
        prvMatchOffsets[prvMatchOffsetsPos] = offs;
      }
      if (offs >= 0xFFFFU)
        throw Plus4Emu::Exception("error in compressed data");
      offs++;
      unsigned int  lzMatchReadAddr = (startAddr - offs) & 0xFFFFU;
      unsigned int  matchLength = readMatchLength();
      for (unsigned int j = 0U; j < matchLength; j++) {
        if (i >= nBytes)                // block should not end within a match
          throw Plus4Emu::Exception("error in compressed data");
        if (!bytesUsed[lzMatchReadAddr])    // byte does not exist yet
          throw Plus4Emu::Exception("error in compressed data");
        if (startAddr > 0xFFFFU || bytesUsed[startAddr])
          throw Plus4Emu::Exception("error in compressed data");
        buf[startAddr] = (buf[lzMatchReadAddr] + deltaValue) & 0xFF;
        bytesUsed[startAddr] = true;
        startAddr++;
        lzMatchReadAddr = (lzMatchReadAddr + 1U) & 0xFFFFU;
        i++;
      }
    }
    return isLastBlock;
  }

  // --------------------------------------------------------------------------

  PRGDecompressor::PRGDecompressor()
    : huffmanLimitTable0((unsigned int *) 0),
      huffmanOffsetTable0((unsigned int *) 0),
      huffmanDecodeTable0((unsigned int *) 0),
      huffmanLimitTable1((unsigned int *) 0),
      huffmanOffsetTable1((unsigned int *) 0),
      huffmanDecodeTable1((unsigned int *) 0),
      usingHuffTable0(false),
      usingHuffTable1(false),
      shiftRegister(0x00),
      shiftRegisterCnt(0),
      inputBuffer((unsigned char *) 0),
      inputBufferSize(0),
      inputBufferPosition(0),
      prvMatchOffsetsPos(0U)
  {
    size_t  totalTableSize = 16 + 16 + 324 + 16 + 16 + 28;
    huffmanLimitTable0 = new unsigned int[totalTableSize];
    for (size_t i = 0; i < totalTableSize; i++)
      huffmanLimitTable0[i] = 0U;
    huffmanOffsetTable0 = &(huffmanLimitTable0[16]);
    huffmanDecodeTable0 = &(huffmanLimitTable0[16 + 16]);
    huffmanLimitTable1 = &(huffmanLimitTable0[16 + 16 + 324]);
    huffmanOffsetTable1 = &(huffmanLimitTable1[16]);
    huffmanDecodeTable1 = &(huffmanLimitTable1[16 + 16]);
    for (size_t i = 0; i < 4; i++)
      prvMatchOffsets[i] = 0xFFFFFFFFU;
  }

  PRGDecompressor::~PRGDecompressor()
  {
    delete[] huffmanLimitTable0;
  }

  void PRGDecompressor::decompressData(
      std::vector< std::vector< unsigned char > >& outBuf,
      const std::vector< unsigned char >& inBuf)
  {
    outBuf.clear();
    if (inBuf.size() < 1)
      return;
    bool    startPosTable[1024];
    for (size_t i = 0; i < 1024; i++)
      startPosTable[i] = false;
    unsigned char crcValue = 0x00;
    // find possible start offsets of compressed data
    for (size_t i = inBuf.size(); i > 0; ) {
      i--;
      crcValue = crcValue ^ inBuf[i];
      crcValue = ((crcValue & 0x7F) << 1) | ((crcValue & 0x80) >> 7);
      crcValue = (crcValue + 0xC4) & 0xFF;
      if (crcValue == 0x80) {
        if (i < 64 || (i >= 576 && i < 1024))
          startPosTable[i] = true;
      }
    }
    std::vector< unsigned char >  tmpBuf;
    std::vector< bool > bytesUsed;
    tmpBuf.resize(65536);
    bytesUsed.resize(65536);
    bool    doneFlag = false;
    for (size_t i = 0; i < 1024; i++) {
      if (!startPosTable[i])
        continue;
      for (size_t j = 0; j < 65536; j++) {
        tmpBuf[j] = 0x00;
        bytesUsed[j] = false;
      }
      // if found a position where the checksum matches, try to decompress data
      inputBuffer = &(inBuf.front());
      inputBufferSize = inBuf.size();
      inputBufferPosition = i + 1;
      shiftRegister = 0x00;
      shiftRegisterCnt = 0;
      try {
        while (!decompressDataBlock(tmpBuf, bytesUsed))
          ;
        // on successful decompression, all input data must be consumed
        if (!(inputBufferPosition >= inputBufferSize && shiftRegister == 0x00))
          throw Plus4Emu::Exception("error in compressed data");
        doneFlag = true;
      }
      catch (Plus4Emu::Exception) {
      }
      if (doneFlag)
        break;
    }
    if (!doneFlag)
      throw Plus4Emu::Exception("error in compressed data");
    unsigned int  startPos = 0U;
    size_t        nBytes = 0;
    // find all data blocks
    for (size_t i = 0; i <= 65536; i++) {
      if (i >= 65536 || !bytesUsed[i]) {
        if (nBytes > 0) {
          std::vector< unsigned char >  newBuf;
          newBuf.push_back((unsigned char) (startPos & 0xFFU));
          newBuf.push_back((unsigned char) ((startPos >> 8) & 0xFFU));
          for (size_t j = 0U; j < nBytes; j++)
            newBuf.push_back(tmpBuf[startPos + j]);
          outBuf.push_back(newBuf);
        }
        startPos = (unsigned int) (i + 1);
        nBytes = 0;
      }
      else {
        nBytes++;
      }
    }
  }

  // --------------------------------------------------------------------------

  const unsigned char PRGDecompressor::sfxModuleLibrary[1904] = {
    0x00, 0x20, 0x17, 0x20, 0x00, 0xDA, 0xD7, 0x62, 0x69, 0x75, 0x56, 0xBD,
    0x25, 0x55, 0xCE, 0xA5, 0xAF, 0xE5, 0xFE, 0xAB, 0x94, 0xFA, 0x5D, 0x45,
    0xBB, 0xD2, 0x4A, 0x73, 0x51, 0xA8, 0xAA, 0x96, 0xFF, 0x67, 0x25, 0xB5,
    0x4D, 0x19, 0xC6, 0x54, 0xBA, 0xD7, 0x2F, 0xD7, 0x3A, 0x95, 0xD6, 0xB4,
    0x05, 0x95, 0x2E, 0x48, 0x5B, 0x22, 0x89, 0x2E, 0x12, 0x9D, 0x47, 0xA6,
    0xDF, 0x29, 0xA4, 0x74, 0xDB, 0x2D, 0xDA, 0x65, 0x24, 0x3E, 0xCE, 0x70,
    0xAF, 0x53, 0x48, 0x55, 0xB9, 0x76, 0xEA, 0x97, 0x4B, 0x56, 0xBC, 0xAD,
    0xEF, 0x5D, 0x2D, 0x4A, 0xD5, 0xE1, 0x2A, 0xDE, 0x5F, 0xAD, 0x2D, 0x4B,
    0x53, 0x9E, 0xDE, 0x71, 0xB9, 0x6A, 0x5A, 0x4B, 0x49, 0x6A, 0x5D, 0xAB,
    0xBD, 0xD8, 0x68, 0x5B, 0x76, 0xA5, 0x72, 0x42, 0x90, 0xA4, 0x99, 0x4D,
    0x2A, 0x0B, 0x72, 0x5A, 0xA4, 0x54, 0x8B, 0x6F, 0xCA, 0xDA, 0xA4, 0xAD,
    0x48, 0x6A, 0x25, 0x1C, 0x68, 0xA9, 0xC9, 0x4A, 0xE5, 0x3F, 0x01, 0x46,
    0x9D, 0xCA, 0x65, 0x73, 0x95, 0x39, 0x69, 0x3D, 0xCE, 0x24, 0xA9, 0x52,
    0xD2, 0x16, 0x92, 0xE8, 0x00, 0x83, 0x5D, 0x80, 0x18, 0x13, 0x10, 0x02,
    0xBD, 0x2D, 0xC0, 0x77, 0xB3, 0x97, 0x21, 0x7B, 0xE0, 0x5D, 0x06, 0xDA,
    0x76, 0x20, 0x86, 0x79, 0x90, 0x59, 0xA8, 0xDE, 0x82, 0x2D, 0x8C, 0x10,
    0xFB, 0xD5, 0x70, 0x02, 0xD1, 0x5D, 0xB8, 0x6C, 0x29, 0x73, 0x99, 0xF4,
    0xAD, 0xEB, 0xE5, 0x77, 0x87, 0x10, 0xD6, 0x78, 0xA4, 0x03, 0x6A, 0x66,
    0x81, 0x37, 0xE3, 0xDF, 0x86, 0x37, 0x90, 0xFC, 0x37, 0xF5, 0x4D, 0xC1,
    0x5F, 0x96, 0xE0, 0x3F, 0x0F, 0x31, 0xD0, 0x30, 0xDA, 0x44, 0x26, 0xBC,
    0xF6, 0x21, 0xF1, 0xED, 0x6D, 0x81, 0x6A, 0xEB, 0xE0, 0xE8, 0xFA, 0x2F,
    0x82, 0xF1, 0x79, 0x21, 0x51, 0x56, 0x79, 0xCD, 0xDB, 0x88, 0xB3, 0xF5,
    0x22, 0x1B, 0x0A, 0xD6, 0x00, 0xFA, 0xBA, 0xD0, 0x16, 0xD7, 0xFC, 0xE0,
    0x3D, 0x87, 0xCA, 0x1F, 0x37, 0xB1, 0x72, 0x1E, 0x86, 0xE2, 0xE8, 0x23,
    0xF7, 0x6F, 0x01, 0xEE, 0xF2, 0x21, 0xE6, 0x91, 0xEE, 0x07, 0x29, 0x8F,
    0xBE, 0x16, 0x16, 0x58, 0x00, 0xAA, 0xCE, 0x00, 0x73, 0xAD, 0x2E, 0x01,
    0x9D, 0xB1, 0xE4, 0xFB, 0x51, 0x1F, 0x77, 0xC1, 0x10, 0xDC, 0x6F, 0x99,
    0x02, 0x12, 0x77, 0xA0, 0xDB, 0xE2, 0xE2, 0x07, 0x9C, 0x57, 0x04, 0x25,
    0x9C, 0xB8, 0x07, 0xF7, 0x4E, 0xB3, 0x23, 0xC7, 0x60, 0x3D, 0xE4, 0x41,
    0xDF, 0xC9, 0x64, 0x0A, 0x38, 0x0D, 0x43, 0x0D, 0xFD, 0xF8, 0x31, 0xE0,
    0xBF, 0x05, 0x77, 0xAD, 0xC3, 0x93, 0x7D, 0xAA, 0x6E, 0x6F, 0xDD, 0x02,
    0x4E, 0x24, 0x43, 0x8B, 0x82, 0xC4, 0x3C, 0x9E, 0x2D, 0xB0, 0x0D, 0x2B,
    0x29, 0x2B, 0x26, 0x82, 0xDE, 0x82, 0x7A, 0x79, 0x4A, 0x0F, 0xEB, 0xD2,
    0xC3, 0x90, 0x51, 0x9B, 0xEB, 0xC8, 0xF3, 0x29, 0xAF, 0x25, 0xFA, 0x61,
    0x5C, 0x68, 0x3E, 0xDC, 0xA9, 0x33, 0x03, 0x41, 0x9B, 0x96, 0x25, 0x44,
    0xA4, 0x1E, 0x7F, 0x50, 0xA0, 0x1E, 0x00, 0xC4, 0xF8, 0x52, 0x58, 0xA5,
    0xC2, 0x67, 0xAA, 0x58, 0xA3, 0x66, 0xB0, 0xC7, 0x24, 0xAA, 0xE6, 0x5F,
    0xDD, 0xC3, 0x99, 0x2A, 0xFE, 0xC4, 0x97, 0xA1, 0x85, 0x6F, 0x4E, 0x62,
    0xA5, 0xEA, 0x48, 0x52, 0xF5, 0x0E, 0xBC, 0x46, 0x12, 0xA6, 0x4F, 0x89,
    0xB5, 0x53, 0xE2, 0x23, 0x93, 0xAC, 0xC1, 0x21, 0x1B, 0x62, 0xFF, 0x77,
    0x36, 0xF8, 0x9D, 0xD1, 0x78, 0xCE, 0xF7, 0x7F, 0x42, 0x32, 0xD0, 0x95,
    0x1D, 0x52, 0x94, 0x6F, 0x76, 0x94, 0xCF, 0x4F, 0xA2, 0x61, 0x43, 0xD8,
    0xD1, 0x22, 0xDD, 0x0E, 0xEA, 0x89, 0xAC, 0xEE, 0x2C, 0xAE, 0x49, 0x40,
    0x89, 0x80, 0xF5, 0x14, 0xCE, 0xF7, 0x56, 0x7E, 0x30, 0xCD, 0x07, 0x44,
    0xBA, 0x53, 0x22, 0x5C, 0xC4, 0x54, 0x23, 0xBD, 0xD6, 0xDA, 0xA7, 0x84,
    0x9E, 0x1F, 0x71, 0x0A, 0x77, 0x41, 0xC8, 0x5E, 0x09, 0x7C, 0xDA, 0xB8,
    0x78, 0x75, 0x3F, 0x68, 0xF0, 0xA7, 0x9A, 0x64, 0x62, 0x52, 0x3D, 0xDA,
    0xE2, 0x48, 0xB4, 0xEE, 0xD8, 0x28, 0x45, 0xC3, 0xD0, 0x45, 0x72, 0xA1,
    0xAA, 0x5F, 0x79, 0x36, 0x84, 0x01, 0xA2, 0xD0, 0x52, 0x06, 0x77, 0x63,
    0x53, 0x98, 0xBD, 0x39, 0x14, 0x0D, 0xBF, 0xC5, 0x3C, 0x45, 0x0B, 0x74,
    0x05, 0x2C, 0x44, 0xA4, 0x7C, 0xF0, 0x87, 0x8B, 0x93, 0xCC, 0x7C, 0x77,
    0x78, 0x6F, 0xA8, 0x35, 0xB5, 0x88, 0x34, 0x74, 0xC5, 0x2F, 0x29, 0xB2,
    0xA8, 0xA5, 0x27, 0x68, 0x1A, 0x2C, 0x36, 0xCE, 0xE6, 0x64, 0x9D, 0x22,
    0x75, 0xA7, 0xB4, 0xF6, 0xD7, 0x11, 0x6B, 0xD9, 0x59, 0xC1, 0xE9, 0x8A,
    0x90, 0xFF, 0x70, 0xB9, 0x8B, 0x22, 0xA1, 0xE5, 0x5B, 0x47, 0x6B, 0xE8,
    0xBA, 0xC8, 0xF2, 0x02, 0x57, 0xE9, 0x9A, 0x7C, 0x33, 0x15, 0xDE, 0xCE,
    0xF2, 0x85, 0x4A, 0x05, 0xA8, 0x36, 0x72, 0x32, 0x48, 0xBD, 0x29, 0x4C,
    0x36, 0x13, 0x3A, 0x46, 0x86, 0x8A, 0xD5, 0xCC, 0xCD, 0x43, 0xDE, 0xDB,
    0x85, 0xF9, 0x8D, 0xC3, 0xE5, 0x8C, 0x4A, 0x41, 0xB3, 0x41, 0x73, 0x6A,
    0xC9, 0x0B, 0xB2, 0xA3, 0x9A, 0xA4, 0x64, 0xCB, 0x85, 0x8C, 0x36, 0x97,
    0x19, 0x68, 0xAB, 0xE5, 0x1D, 0x4C, 0xBB, 0x06, 0xAB, 0x5B, 0x35, 0xAA,
    0x76, 0x61, 0x14, 0x72, 0x64, 0x2B, 0x72, 0x97, 0x29, 0x6B, 0x55, 0xCE,
    0x66, 0xB2, 0xAA, 0x9A, 0xD6, 0xA2, 0x72, 0x39, 0x14, 0x98, 0x83, 0x39,
    0xEC, 0x92, 0x23, 0x65, 0x85, 0x48, 0x5C, 0x76, 0x16, 0xCA, 0x1C, 0xD6,
    0x5A, 0xD2, 0xC3, 0x2E, 0xAB, 0xAF, 0x25, 0xF4, 0x79, 0xE0, 0x3C, 0x5E,
    0xAF, 0x3D, 0x96, 0xA9, 0x61, 0x4D, 0x74, 0x87, 0x34, 0x71, 0x12, 0x58,
    0x61, 0x85, 0x35, 0xD2, 0x0B, 0xE3, 0xE3, 0xD9, 0x4E, 0x77, 0x30, 0xD9,
    0xD3, 0xC8, 0x48, 0xC0, 0x59, 0x27, 0xA3, 0xE4, 0x0F, 0x29, 0x34, 0xD9,
    0x7B, 0xF8, 0x6D, 0x8A, 0x31, 0xD8, 0xCC, 0x14, 0x63, 0xC7, 0xA3, 0x86,
    0xA5, 0x33, 0xDE, 0xF3, 0xEC, 0x5A, 0xF7, 0x25, 0xA7, 0x7D, 0xD5, 0xA2,
    0xD0, 0xB7, 0x40, 0x8A, 0xD4, 0x74, 0xC9, 0x54, 0x27, 0x93, 0xCD, 0x0B,
    0xB7, 0x2A, 0x3B, 0x3A, 0x36, 0x7B, 0x8D, 0xD6, 0xCA, 0xF5, 0xC4, 0x4F,
    0xF5, 0x45, 0x7D, 0x69, 0x0A, 0x29, 0xC3, 0x8F, 0x3F, 0xD4, 0x1F, 0x4F,
    0xAE, 0xCE, 0xC8, 0xBC, 0x12, 0xA4, 0xB5, 0xE5, 0x48, 0xFF, 0x62, 0x16,
    0xDD, 0x84, 0x87, 0xEC, 0x08, 0xFF, 0xBA, 0x51, 0x51, 0x31, 0x09, 0x49,
    0x61, 0x09, 0x53, 0x94, 0x89, 0x88, 0x46, 0xF9, 0x49, 0x09, 0x1C, 0xCC,
    0x2E, 0xD2, 0xD4, 0xB3, 0x39, 0xEC, 0xD3, 0xF8, 0x8B, 0x3C, 0xA5, 0x58,
    0xD3, 0xF4, 0xA6, 0xA8, 0xA8, 0x79, 0x8B, 0x57, 0x38, 0x3C, 0x94, 0x76,
    0x6B, 0x65, 0x1A, 0x08, 0x73, 0xD7, 0x15, 0x6F, 0x07, 0x6D, 0x24, 0xD7,
    0x5D, 0x98, 0x3A, 0x0D, 0xE3, 0xAF, 0x15, 0xA1, 0x25, 0x04, 0x8E, 0x39,
    0x7F, 0xCE, 0x46, 0x30, 0xEC, 0x8B, 0xC9, 0xAB, 0xB3, 0x97, 0x33, 0xA6,
    0x47, 0x98, 0xCE, 0x94, 0xCD, 0xA5, 0x4E, 0xE6, 0xC4, 0xD7, 0xF6, 0x4F,
    0x44, 0x6B, 0x74, 0xED, 0x01, 0x0A, 0x5D, 0xAD, 0x21, 0x51, 0xF6, 0x2B,
    0x0E, 0x09, 0x06, 0xBB, 0x94, 0x92, 0xC4, 0xEC, 0x44, 0x7E, 0xA2, 0xBC,
    0xFC, 0x50, 0xFC, 0x54, 0x1B, 0x31, 0x78, 0x35, 0xC7, 0x07, 0x38, 0x47,
    0x18, 0x5D, 0x47, 0x22, 0xFE, 0x8C, 0x05, 0xF7, 0x45, 0xEB, 0x0B, 0x9F,
    0x94, 0x23, 0xF1, 0x17, 0x07, 0xE4, 0x27, 0xE1, 0x60, 0x9A, 0xD2, 0xE9,
    0xCD, 0x9B, 0xBD, 0x53, 0x58, 0x4F, 0x6E, 0x7C, 0xF2, 0x05, 0xF2, 0x6A,
    0xED, 0x9B, 0x88, 0xEF, 0xCA, 0x6E, 0x2B, 0xE6, 0xA7, 0x3A, 0x02, 0x9C,
    0x81, 0x04, 0x23, 0xA9, 0x48, 0x31, 0xCB, 0x38, 0xFF, 0x51, 0xA8, 0x37,
    0x5E, 0x9F, 0xFA, 0x5E, 0x9F, 0xDF, 0x6E, 0x05, 0xC9, 0x70, 0x2E, 0xEE,
    0xF9, 0x48, 0x91, 0xA7, 0x33, 0x1C, 0xB3, 0x95, 0x70, 0xFD, 0x73, 0x52,
    0x4E, 0xB0, 0xF9, 0xE4, 0x8A, 0xE3, 0x81, 0x43, 0x81, 0x75, 0x0E, 0x4A,
    0x93, 0x91, 0x1F, 0x84, 0xFC, 0x2C, 0x58, 0xCB, 0x9B, 0x87, 0x24, 0x2D,
    0x3F, 0x9B, 0x1E, 0xBF, 0x61, 0x5E, 0x74, 0x28, 0x75, 0x19, 0x6B, 0x5E,
    0x27, 0xE3, 0x5B, 0xB6, 0x46, 0x50, 0x50, 0x38, 0x35, 0x34, 0x38, 0x3D,
    0xF8, 0xE5, 0xB7, 0xE3, 0x51, 0xE8, 0xCE, 0x42, 0xF6, 0xDB, 0x9A, 0x6E,
    0x64, 0x8F, 0x9F, 0xAF, 0x4F, 0x9A, 0xE3, 0x11, 0x43, 0x18, 0xC8, 0xB7,
    0xE1, 0x33, 0x0A, 0x82, 0x31, 0xDB, 0x3D, 0x6C, 0x68, 0xF8, 0xB6, 0x3D,
    0xF8, 0xE1, 0xA7, 0xC1, 0x50, 0xEA, 0x23, 0xF9, 0x0B, 0xED, 0x3C, 0x28,
    0xBE, 0x8C, 0x69, 0x7E, 0x13, 0xF6, 0xAE, 0x9A, 0x31, 0xEF, 0x32, 0xF8,
    0xD7, 0x10, 0xF8, 0xF7, 0xE3, 0x98, 0x1F, 0xBF, 0xFE, 0xBD, 0x7E, 0xA2,
    0x33, 0x93, 0x27, 0x4F, 0xF6, 0x09, 0x8C, 0x8D, 0x7E, 0x13, 0xF6, 0xAD,
    0x0C, 0x64, 0xFE, 0x5F, 0x53, 0xBB, 0xB1, 0xEF, 0xC7, 0x1F, 0x3F, 0x0A,
    0xAB, 0x51, 0x13, 0xC7, 0xED, 0x4D, 0x95, 0x71, 0x95, 0x2F, 0xC6, 0xB4,
    0x7B, 0x8C, 0x94, 0xCE, 0xB7, 0x35, 0x1C, 0x37, 0x3D, 0xF8, 0xE5, 0x5D,
    0x42, 0x29, 0x4B, 0xDA, 0x4F, 0x29, 0x5D, 0x5C, 0xA9, 0x7E, 0xA5, 0x75,
    0xF2, 0xA5, 0xD3, 0x57, 0x75, 0x3B, 0xDE, 0xBA, 0xA9, 0xB5, 0x75, 0x47,
    0x6F, 0xA5, 0x73, 0xEB, 0xD3, 0x19, 0x33, 0xCA, 0xD1, 0xED, 0x94, 0xD9,
    0xDD, 0x6B, 0x2B, 0xD5, 0xEF, 0x91, 0x17, 0x14, 0xBF, 0x77, 0xEA, 0xBE,
    0x33, 0xE5, 0x7D, 0x28, 0x01, 0x9F, 0x44, 0xD6, 0xE7, 0xC3, 0x7E, 0xA9,
    0xBA, 0x76, 0xA7, 0xCD, 0x34, 0xF8, 0x00, 0x10, 0x16, 0xE8, 0x6D, 0x8B,
    0xB2, 0xDD, 0x97, 0x49, 0xFD, 0xE4, 0xCA, 0xBA, 0x4D, 0x59, 0x78, 0xC0,
    0xC9, 0x8E, 0x1E, 0x61, 0x4B, 0xE9, 0xE7, 0x41, 0xF3, 0x78, 0xCA, 0xE6,
    0x71, 0x25, 0xD7, 0x93, 0xDD, 0x91, 0x3F, 0x64, 0x9D, 0x32, 0x36, 0x87,
    0x53, 0x6C, 0x3C, 0xCB, 0xEF, 0x75, 0x9C, 0x3D, 0x27, 0x42, 0xF0, 0xA8,
    0x97, 0xF4, 0xA0, 0xA2, 0xDE, 0x4F, 0x76, 0x44, 0xBA, 0x85, 0x45, 0x8D,
    0x48, 0x47, 0x29, 0x87, 0x99, 0x09, 0xE5, 0xBE, 0xE8, 0x7D, 0xCA, 0xA5,
    0xED, 0x2A, 0x2B, 0x1D, 0xF4, 0xF0, 0x6C, 0x61, 0x4E, 0x67, 0xD2, 0x12,
    0xCA, 0x61, 0xE6, 0x59, 0x78, 0x22, 0xD7, 0x74, 0x4E, 0xC5, 0xE9, 0x71,
    0xCB, 0xC5, 0x0E, 0xF2, 0x65, 0x5D, 0x27, 0x2C, 0xDC, 0x80, 0x74, 0xC9,
    0x0F, 0x30, 0xC3, 0x8B, 0x41, 0xD1, 0x7C, 0x0E, 0x12, 0xBF, 0x9C, 0x09,
    0x95, 0xE4, 0xF4, 0xFA, 0xC4, 0xEF, 0x91, 0xB0, 0x37, 0x9B, 0x21, 0xE6,
    0x5E, 0x7B, 0xA7, 0x47, 0x7D, 0xD1, 0x1C, 0x2C, 0xF2, 0xFA, 0x99, 0xCA,
    0x25, 0xE4, 0xF4, 0xF4, 0x4A, 0x6F, 0x19, 0xFB, 0x2A, 0x87, 0x98, 0xF1,
    0xE5, 0xB6, 0xE8, 0x5D, 0xCF, 0xF9, 0x97, 0x58, 0xA9, 0x5E, 0x0C, 0xDB,
    0x29, 0x0C, 0xCA, 0x02, 0x29, 0x42, 0xD0, 0x65, 0x37, 0x81, 0xF7, 0x9D,
    0x05, 0xB9, 0xF1, 0x2D, 0xA9, 0xF0, 0x4F, 0x6F, 0x26, 0x55, 0xD2, 0x66,
    0xCA, 0xC4, 0x05, 0xCC, 0x50, 0xF3, 0x09, 0x38, 0xB3, 0x5C, 0xFF, 0xE3,
    0xF8, 0xAB, 0xCA, 0xE6, 0x8F, 0x45, 0xC7, 0x9C, 0xAD, 0xA7, 0x32, 0x59,
    0x79, 0x3D, 0x5A, 0x1F, 0xF6, 0xBB, 0xF8, 0xE1, 0xF7, 0xFD, 0xC2, 0xCA,
    0x3F, 0xF7, 0xD1, 0x3A, 0xD7, 0x80, 0x71, 0xFD, 0x1C, 0x77, 0x87, 0xD3,
    0x7C, 0xDE, 0x66, 0x27, 0x83, 0xBD, 0xD0, 0xDF, 0xCB, 0xA1, 0xB6, 0xAB,
    0x1A, 0x0A, 0x29, 0x17, 0x93, 0xD3, 0xC9, 0x65, 0xBF, 0xC0, 0xAB, 0x6A,
    0xE4, 0x45, 0xF2, 0xD2, 0x13, 0xCD, 0x30, 0xF3, 0x25, 0x3C, 0xBA, 0x6E,
    0x8D, 0xDC, 0xD7, 0xFC, 0x02, 0x9A, 0xCB, 0x05, 0xE4, 0xF0, 0x75, 0x9A,
    0x14, 0xF9, 0x74, 0x8A, 0xFD, 0x6D, 0xEB, 0x7C, 0xAB, 0x1E, 0x09, 0x89,
    0x56, 0x1D, 0x29, 0xB1, 0x82, 0x5D, 0xF3, 0x01, 0x4B, 0xBC, 0x9E, 0x02,
    0xB4, 0x5D, 0x30, 0xB1, 0x98, 0x12, 0x99, 0x8A, 0x09, 0xE6, 0x20, 0x71,
    0x6A, 0x59, 0x2E, 0xE5, 0x13, 0xA8, 0x0E, 0x9B, 0x80, 0x1F, 0xD5, 0xA4,
    0x33, 0x30, 0x1D, 0x3D, 0x20, 0x3A, 0x63, 0x7E, 0xC9, 0x77, 0x78, 0xA6,
    0x5E, 0x9E, 0xF0, 0x41, 0x53, 0x1D, 0x91, 0x00, 0x40, 0x5A, 0xF3, 0x2D,
    0xFB, 0x50, 0xDD, 0xA0, 0x9B, 0xCD, 0x92, 0xEE, 0x35, 0xA0, 0x04, 0x60,
    0x43, 0xC5, 0x16, 0xF2, 0xBE, 0x03, 0xDF, 0xB5, 0xE6, 0x1A, 0x79, 0x7B,
    0x27, 0x47, 0xF8, 0xFD, 0xAA, 0xF3, 0xB9, 0xA3, 0xD5, 0x71, 0xE5, 0x2C,
    0x29, 0xC8, 0x9A, 0x5E, 0x4F, 0x76, 0x4A, 0xFD, 0x96, 0x54, 0xFB, 0x3E,
    0x4A, 0x02, 0xBC, 0x6E, 0x68, 0xF6, 0x5C, 0x77, 0x07, 0xB3, 0x74, 0x3C,
    0xCC, 0xAF, 0x75, 0xA1, 0x7D, 0x27, 0x48, 0xF1, 0x6B, 0x1D, 0x1A, 0x29,
    0x77, 0x93, 0xD3, 0xC8, 0x65, 0x9F, 0xC0, 0xAB, 0xAB, 0xBA, 0x22, 0xF9,
    0x54, 0x13, 0x4A, 0xA1, 0xE6, 0x48, 0x79, 0x6F, 0xBA, 0x4F, 0x7F, 0x54,
    0xBF, 0xE6, 0xA2, 0xBD, 0x79, 0x3D, 0xD5, 0x45, 0xEA, 0x3A, 0x53, 0x97,
    0x5B, 0x92, 0x6B, 0x32, 0xFB, 0xC1, 0x2B, 0xAA, 0x3A, 0x57, 0x7E, 0xFC,
    0xBB, 0x25, 0xF2, 0x93, 0x79, 0x3C, 0x04, 0xE8, 0x3A, 0x5B, 0x63, 0x28,
    0x23, 0x32, 0x93, 0xF3, 0xCC, 0x3C, 0xE2, 0xD3, 0xB2, 0x67, 0xCA, 0x27,
    0x10, 0x1D, 0x2F, 0x00, 0x3E, 0xAB, 0x47, 0x66, 0x40, 0x3A, 0x78, 0xC0,
    0x74, 0xA6, 0xFD, 0x93, 0x3E, 0xF1, 0x4A, 0xBD, 0x3D, 0xD0, 0x7E, 0xA6,
    0x2B, 0x21, 0x00, 0x80, 0xB5, 0xE6, 0x59, 0xF6, 0xA1, 0xB3, 0x41, 0x2F,
    0x9B, 0x26, 0x7C, 0x4B, 0x40, 0x08, 0x80, 0x87, 0x8A, 0x1D, 0xE5, 0x7C,
    0x06, 0xBF, 0x6B, 0xCC, 0x30, 0xF2, 0xF4, 0xE0
  };

  void PRGDecompressor::getSFXModule(std::vector< unsigned char >& outBuf,
                                     int runAddr, bool c16Mode,
                                     bool fastSFXModule, bool noCleanup,
                                     bool noROM, bool noCLI)
  {
    bool    isBasicProgram = (runAddr == -1);
    if (runAddr < -3)
      runAddr = -2;
    else if (runAddr > 0xFFFF)
      runAddr = runAddr & 0xFFFF;
    outBuf.clear();
    int     decompressModuleIndex =
        (c16Mode ? 1 : 0) | (fastSFXModule ? 2 : 0) | (noCleanup ? 4 : 0)
        | (noROM ? 8 : 0) | (noCLI ? 16 : 0) | (isBasicProgram ? 32 : 0);
    std::vector< unsigned char >  tmpBuf1;
    std::vector< std::vector< unsigned char > > tmpBuf2;
    for (size_t i = 0; i < sizeof(sfxModuleLibrary); i++)
      tmpBuf1.push_back(sfxModuleLibrary[i]);
    PRGDecompressor decompressor;
    decompressor.decompressData(tmpBuf2, tmpBuf1);
    unsigned int  baseAddr =
        (unsigned int) tmpBuf2[0][0] | ((unsigned int) tmpBuf2[0][1] << 8);
    unsigned char *p = &(tmpBuf2[0].front()) + (decompressModuleIndex * 4 + 2);
    unsigned int  offs =
        ((unsigned int) p[0] | ((unsigned int) p[1] << 8)) - baseAddr;
    unsigned int  len = (unsigned int) p[2] | ((unsigned int) p[3] << 8);
    p = &(tmpBuf2[0].front()) + (offs + 2U);
    for (unsigned int i = 0U; i < len; i++)
      outBuf.push_back(p[i]);
    if (!isBasicProgram) {
      if (runAddr == -3)
        runAddr = 0xFF52;
      else if (runAddr < 0)
        runAddr = 0x867E;
      outBuf[outBuf.size() - 2] = (unsigned char) (runAddr & 0xFF);
      outBuf[outBuf.size() - 1] = (unsigned char) ((runAddr >> 8) & 0xFF);
    }
  }

}       // namespace Plus4FLIConv

